/*-----------------------------------------------------------------------------------------------------------*
  System includes 
*------------------------------------------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/types.h>
#include <nn/os.h>
#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/hid.h>
#include <nn/fnd.h>
#include <nn/y2r.h>
#include <nn/camera.h>
#include <nn/applet.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

/*-----------------------------------------------------------------------------------------------------------*
  User includes
*------------------------------------------------------------------------------------------------------------*/

#include "mocam_Autofocus.h"

#include "helper_DisplayDevice.h"
#include "helper_RgbImage.h"
#include "helper_YuvToRgb.h"
#include "helper_Camera.h"
#include "helper_Font.h"
#include "helper_Application.h"

/*-----------------------------------------------------------------------------------------------------------*
  Constants definition.
*------------------------------------------------------------------------------------------------------------*/

const size_t                        ROMFS_BUFFER_SIZE    = (1024 * 64);

// for 512x384 at 20fps capture
const int                           CAMERA_WIDTH        = 512;
const int                           CAMERA_HEIGHT       = 384;
const nn::camera::CTR::Size         CAMERA_RESOLUTION   = nn::camera::CTR::SIZE_DS_LCDx4;
const nn::camera::CTR::FrameRate    CAMERA_FRAMERATE    = nn::camera::CTR::FRAME_RATE_20;
nn::camera::CTR::CameraSelect       CAMERA_SELECT       = nn::camera::CTR::SELECT_OUT1_OUT2;

/*-----------------------------------------------------------------------------------------------------------*
  Application global variables used by application
*------------------------------------------------------------------------------------------------------------*/

static nn::fnd::ExpHeap             s_Heap;
static nn::fnd::ExpHeap::Allocator  s_HeapAllocator(s_Heap);
static char                         s_RomBuffer[ROMFS_BUFFER_SIZE];

/*-----------------------------------------------------------------------------------------------------------*
  Allocator functions that will be provided to the mocam library
*------------------------------------------------------------------------------------------------------------*/

static void* Malloc(size_t size)  { return s_Heap.Allocate(size); }
static void  Free(void* ptr)      { if (ptr) s_Heap.Free(ptr); }

/*-----------------------------------------------------------------------------------------------------------*
  Small mathematical functions used by the application
*------------------------------------------------------------------------------------------------------------*/

inline int32_t Round(float f)
{
    return (f<0?(int32_t)(f-0.5f):(int32_t)(f+0.5f));
}

/*-----------------------------------------------------------------------------------------------------------*/

template <typename T> 
T Min(const T& a,const T& b)
{
    return (a<b?a:b);
}

/*-----------------------------------------------------------------------------------------------------------*/

template <typename T> 
T Max(const T& a,const T& b)
{
    return (a>b?a:b);
}

/*-----------------------------------------------------------------------------------------------------------*
  AutoFocus callback
*------------------------------------------------------------------------------------------------------------*/

static void AutoFocusCallback(void* param)
{
    // Give some time to the main thread
    // A good way to handle this should be to use Yield instead of Sleep
    // but the Yield doesn't seem to work in SDK 0.10.2
    u32 sleepTime=*(u32*)param;
    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(sleepTime));
}

/*-----------------------------------------------------------------------------------------------------------*
  StereoAutofocus class
*------------------------------------------------------------------------------------------------------------*/

class StereoAutofocus
{
private:
    mw::mo::mocam::autofocus::Handle    Autofocus;
    u8*                                 Image[2];
    nn::os::LightEvent                  Event[2];
    float                               TranslationX;
    float                               TranslationY;

public:
    u32                                 SleepTime;

public:
    StereoAutofocus():TranslationX(0.f),TranslationY(0.f),SleepTime(30)
    {
        Image[0]=0;
        Image[1]=0;
        for (u32 i=0; i<2; i++)
            Event[i].Initialize(false);
    }
    ~StereoAutofocus()
    {
        Finalize();        
        for (u32 i=0; i<2; i++)
            Event[i].Finalize();
    }
    bool Initialize()
    {
        if (Autofocus.Initialize(Malloc,Free)!=mw::mo::mocam::OK)
        {
            Finalize();
            return false;
        }
        Image[0]=(u8*)Malloc(CAMERA_WIDTH*CAMERA_HEIGHT*2);
        Image[1]=(u8*)Malloc(CAMERA_WIDTH*CAMERA_HEIGHT*2);
        if (!Image[0] || !Image[1])
        {
            Finalize();
            return false;
        }
        // Initialize autofocus callback
        Autofocus.SetCallback(AutoFocusCallback,&SleepTime);
        return true;
    }
    void Finalize()
    {
        Autofocus.Finalize();
        Free(Image[0]);
        Image[0]=0;
        Free(Image[1]);
        Image[1]=0;
    }
    void InitImages(const u8* left,const u8* right)
    {
        memcpy(Image[0],left,CAMERA_WIDTH*CAMERA_HEIGHT*2);
        memcpy(Image[1],right,CAMERA_WIDTH*CAMERA_HEIGHT*2);
    }
    bool IsReadyForNextFrame()
    {
        return Event[0].TryWait();
    }
    void Synchronize()
    {
        Event[0].Wait();
    }
    void NextFrame()
    {
        return Event[1].Signal();
    }
    void SetTranslation(float tx,float ty)
    {
        TranslationX=tx;
        TranslationY=ty;
    }
    void SetFocus(float focus)
    {
        Autofocus.Set(focus);
    }
    void Process()
    {
        for(;;)
        {
            // Signal that autofocus is ready for the next frame
            Event[0].Signal();
            
            // Wait for the next frame
            Event[1].Wait();

            // Use TranslationX and TranslationY to "calibrate" source pictures
            s32 tx=Round(TranslationX);
            s32 ty=Round(TranslationY);
            const u8* src0=&Image[0][(tx<0?-tx*2:0)+(ty>0?ty:0)*CAMERA_WIDTH*2];
            const u8* src1=&Image[1][(tx>0?tx*2:0)+(ty<0?-ty:0)*CAMERA_WIDTH*2];
            u32 width=CAMERA_WIDTH-abs(tx);
            u32 height=CAMERA_HEIGHT-abs(ty);

            // Calculate Autofocus internal value
            Autofocus.Calculate(src0,src1,width,height,CAMERA_WIDTH*2,width-400);
        }
    }
    f32 GetValue()
    {
        // Get current interpolated focus value
        f32 focus;        
        Autofocus.Get(1.5f,focus);
        return focus;
    }
    f32 GetTranslationX()
    {
        return TranslationX;
    }
    f32 GetTranslationY() 
    {
        return TranslationY;
    }
};

/*-----------------------------------------------------------------------------------------------------------*
  AutoFocus thread
*------------------------------------------------------------------------------------------------------------*/

static void AutoFocusThread(StereoAutofocus* autofocus)
{    
    autofocus->Process();
}

/*-----------------------------------------------------------------------------------------------------------*
  Fractal class (used to simulate a CPU intensive process)
*------------------------------------------------------------------------------------------------------------*/

class Fractal
{
private:
    u8*      Buffer[3];
    u32      Index;
    u32      Width;
    u32      Height;
    float    MinRe;
    float    MaxRe;
    float    MinIm;
    float    MaxIm;
    float    IncRe;
    float    IncIm;
    float    PosIm;
    float    Ca;
    float    Cb;
    float    IncCa;
    float    IncCb;
    u32      Line;
    u32      Iteration;
    u32      TotalCount;
    u32      CurrentCount;
    u32      StepByCall;    

public:
    Fractal()
    {
        Buffer[0]=Buffer[1]=Buffer[2]=0;
    }

    ~Fractal()
    {
        Finalize();
    }

    void InitStep()
    {        
        Line=0;
        PosIm=MaxIm;
        memset(Buffer[(Index+1)%3],255,Width*Height*3);
    }

    bool Initialize(u32 stepByCall)
    {
        StepByCall=stepByCall;
        Index=0;
        Width=320;
        Height=240;
        MinRe=-2.f;
        MaxRe=2.f;
        MinIm=-1.5f;
        MaxIm=MinIm+(MaxRe-MinRe)*(float)Height/(float)Width;
        IncRe=(MaxRe-MinRe)/((float)Width-1);
        IncIm=(MaxIm-MinIm)/((float)Height-1);
        TotalCount=10;
        CurrentCount=0;
        Iteration=10;    
        Ca=0.353f;
        Cb=0.288f;
        srand(0);
        IncCa=-0.05f+(float)(rand()%101)/1000.f;
        IncCb=-0.05f+(float)(rand()%101)/1000.f;
        for (u32 i=0; i<3; i++)
        {
            Buffer[i]=(u8*)Malloc(Width*Height*3);
            if (!Buffer[i])
                return false;
        }
        InitStep();
        return true;
    }

    void Finalize()
    {
        for (u32 i=0; i<3; i++)
        {
            Free(Buffer[i]);
            Buffer[i]=0;
        }
    }

    const u8* GetBuffer() const
    {
        return Buffer[Index];
    }

    void Step()
    {
        u8* dst=Buffer[(Index+1)%3];
        float posRe=MinRe;
        for (u32 x=0; x<Width; x++)
        {            
            float zRe=posRe;
            float zIm=PosIm;
            u32 n=0;
            for (; n<Iteration; n++)
            {
                float zRe2=zRe*zRe;
                float zIm2=zIm*zIm;
                if (zRe2+zIm2>4.f)
                    break;
                zIm=2*zRe*zIm+Ca;
                zRe=zRe2-zIm2+Cb;
            }
            u32 r,g,b;
            if (n!=Iteration)
            {                
                if (n<=(Iteration>>1))
                {
                    r=(n*255)/(Iteration>>1);
                    g=0;
                    b=0;
                }
                else
                {
                    n=((n-(Iteration>>1))*255)/(Iteration>>1);
                    r=n;
                    g=n;
                    b=n;
                }
                dst[(Height-Line-1)*3+0+x*Height*3]=b;
                dst[(Height-Line-1)*3+1+x*Height*3]=g;
                dst[(Height-Line-1)*3+2+x*Height*3]=r;
            }
            posRe+=IncRe;
        }
        PosIm-=IncIm;
        Line++;
        if (Line==Height)
        {
            Index=(Index+1)%3;
            CurrentCount++;
            if (CurrentCount==TotalCount)
            {
                IncCa=-0.05f+(float)(rand()%101)/1000.f;
                IncCb=-0.05f+(float)(rand()%101)/1000.f;
                if ((Ca<=-0.9f && IncCa<0.f) || (Ca>=0.9f && IncCa>0.f))
                    IncCa=-IncCa;
                if ((Cb<=-0.9f && IncCb<0.f) || (Cb>=0.9f && IncCb>0.f))
                    IncCb=-IncCb;
                CurrentCount=0;
            }
            Ca+=IncCa;
            Cb+=IncCb;
            InitStep();
        }
    }    

    void Process()
    {
        for (u32 i=0; i<StepByCall; i++)
            Step();
    }
};

/*-----------------------------------------------------------------------------------------------------------*
  Callback that is called just before entering into SLEEP state or in the HOME menu.
*------------------------------------------------------------------------------------------------------------*/

void BeforeEventCallback(void* pUserData, mw::mo::helper::application::EventType eventType)
{
    (void)eventType;

    mw::mo::helper::Camera* camera = (mw::mo::helper::Camera*)pUserData;
    if (camera != NULL)
        camera->StopCapture();
}

/*-----------------------------------------------------------------------------------------------------------*
  Callback that is called just after leaving the SLEEP state or the HOME menu.  
*------------------------------------------------------------------------------------------------------------*/

void AfterEventCallback(void* pUserData, mw::mo::helper::application::EventType eventType)
{
    if (eventType == mw::mo::helper::application::EVENT_SLEEP)
        nn::gx::StartLcdDisplay();
    else
    {
        nngxUpdateState(NN_GX_STATE_ALL);
        nngxValidateState(NN_GX_STATE_ALL,GL_TRUE);
        glGetError(); //nngxValidateState can generate errors when returning back from HOME menu
    }

    mw::mo::helper::Camera* camera = (mw::mo::helper::Camera*)pUserData;
    if (camera != NULL)
        camera->StartCapture();
}

/*-----------------------------------------------------------------------------------------------------------*
  The main function
*------------------------------------------------------------------------------------------------------------*/

void nnMain ()
{
    /*-------------------------------------------------------------------------------------------------------*
      Various system inits
     *-------------------------------------------------------------------------------------------------------*/
    
    mw::mo::helper::application::Initialize();
    nn::applet::Enable();

    if (mw::mo::helper::application::ProcessCloseStatus() == true)
    {
        NN_LOG("Application ended...\n");
        mw::mo::helper::application::Finalize();
        nn::applet::PrepareToCloseApplication();
        nn::applet::CloseApplication();        
    }

    nn::fs::Initialize();
    nn::hid::Initialize();

    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, s_RomBuffer, ROMFS_BUFFER_SIZE));

    s_Heap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize(), nn::os::ALLOCATE_OPTION_LINEAR);

    mw::mo::helper::yuvtorgb::Initialize();

    /*-------------------------------------------------------------------------------------------------------*
      Initializes all helper modules that will be used.
     *-------------------------------------------------------------------------------------------------------*/

    // Display initialization
    mw::mo::helper::displaydevice::Initialize(&s_HeapAllocator);
    
    // Camera initialization 
    mw::mo::helper::Camera camera;
    camera.Initialize(&s_HeapAllocator,CAMERA_SELECT,CAMERA_RESOLUTION,CAMERA_FRAMERATE);

    // Parallax offsets used to display left and right camera images
    int parallax[2]={(CAMERA_WIDTH-400)/2,(CAMERA_WIDTH-400)/2};

    // Display buffers initialization
    u8*                         imageData[2];
    mw::mo::helper::RgbImage    image[2];
    for (u32 i=0; i<2; i++)
    {
        imageData[i]=(u8*)s_Heap.Allocate(512*512*3,128);
        image[i].Initialize(512,512,parallax[i],(CAMERA_HEIGHT-240)/2,400,240,GL_RGB_NATIVE_DMP,GL_UNSIGNED_BYTE,imageData[i]);
    }

    // Start camera capture
    camera.StartCapture();

    /* Configures application callbacks for HOME and SLEEP managment. */
    mw::mo::helper::application::SetCallbacks(BeforeEventCallback, AfterEventCallback);
    mw::mo::helper::application::SetUserData(&camera);
    
    /*-------------------------------------------------------------------------------------------------------*
      Autofocus initialization and configuration.
     *-------------------------------------------------------------------------------------------------------*/

    nn::camera::StereoCameraCalibrationData calibrationData;

    // We retrieve the nn::camera::StereoCameraCalibrationData values of the CTR device.
    nn::camera::GetStereoCameraCalibrationData(&calibrationData);

    // StereoAutofocus class initialization
    StereoAutofocus autofocus;

    if (!autofocus.Initialize())
    {
        NN_LOG("Unable to initialize autofocus !\n");
        nn::dbg::Panic();
    }
    
    // Compute the translation parameters that will be given to StereoAutofocus class.
    //
    // These translation parameters ('txInfinite' for horizontal translation and 'tyInfinite' for vertical translation) are 
    // 'txInfinite' and 'tyInfinite' are the translation amount in pixels that we have to apply to left camera image to make it 
    // match the right camera image match at infinite distance (far away objects).
    //
    // 'translationX' and 'translationX' fields of nn::camera::StereoCameraCalibrationData structure store
    // the translation amount in pixels that we have to apply to left camera image to make it 
    // match the right camera image match at a distance of 'distanceToChart' field of 
    // nn::camera::StereoCameraCalibrationData structure (value is always 250 millimeters).
    // As a consequence we'll have to compute 'txInfinite' value using various fields value of 
    // nn::camera::StereoCameraCalibrationData structure.

    // Compute the average of fov values in degrees of left and right cameras, 
    // just in case they have different values (should be 66 degrees for both) 
    f32 fov = (calibrationData.angleOfViewLeft + calibrationData.angleOfViewRight)/2.f;
    // Compute camera calibration distance, in meters
    f32 calibrationDistance = calibrationData.distanceToChart/1000.f;                   

    // Compute the 'txInfinite' value.
    // For a better understanding of the computation we're doing here, please have a look at 
    // the 'mw::mo::mocam::DistanceToParallax' description in the CTR Mocam SDK Function Reference as
    // the following equation is shown: T25x=Tx*(512/640)+DistanceToParallax(512,66,0.25)
    // Here we modify the equation to obtain 'Tx' so we have:
    // Tx=(T25x-DistanceToParallax(512,66,0.25))/(512/640)
    f32 txInfinite = (calibrationData.translationX - mw::mo::mocam::DistanceToParallax(calibrationData.imageWidth, fov, calibrationDistance)) / 
                     ((float)CAMERA_WIDTH / calibrationData.imageWidth);

    // 'tyInfinite' does not depend on distance of the calibration chart.
    f32 tyInfinite = calibrationData.translationY * ((float)CAMERA_HEIGHT / calibrationData.imageHeight);

    // Give the 'txInfinite' and 'tyInfinite' parameters to StereoAutofocus class.
    autofocus.SetTranslation(txInfinite, tyInfinite);

    // Autofocus thread initialization
    if (!autofocus.Initialize())
    {
        NN_LOG("Unable to initialize autofocus !\n");
        nn::dbg::Panic();
    }
    nn::os::Thread autofocusThread;
    autofocusThread.StartUsingAutoStack(AutoFocusThread,&autofocus,4096,16);

    /*-------------------------------------------------------------------------------------------------------*
      Fractal initialization.
     *-------------------------------------------------------------------------------------------------------*/

    Fractal fractal;
    if (!fractal.Initialize(50))
    {
        NN_LOG("Cannot initialize fractal\n");
        nn::dbg::Panic();
    }

    /*-------------------------------------------------------------------------------------------------------*
      A button:                enable/disable 3D autofocus
      Left/Right buttons:    decrease/increase 3D autofocus thread sleeping time
     *-------------------------------------------------------------------------------------------------------*/

    nn::hid::CTR::PadReader padReader;

    // Autofocus is enabled when the application starts.
    // Use A button to switch it on/off
    bool autoFocusEnable=true;

    // Used to count how many autofocus called are made per camera captured frames
    u32 loopCount=0;
    u32 autofocusCount=0;    

    // Used to count frame per seconds
    nn::os::Tick startTime=nn::os::Tick::GetSystemCurrent();
    u32 frameCount=0;

    /*-------------------------------------------------------------------------------------------------------*
      The main loop
     *-------------------------------------------------------------------------------------------------------*/

    for(;;)
    {
        nn::hid::CTR::PadStatus padStatus;
        if (padReader.ReadLatest(&padStatus))
        {
            // A button enable/disable autofocus
            if (padStatus.trigger & nn::hid::CTR::BUTTON_A)
            {
                autoFocusEnable=!autoFocusEnable;
                loopCount=0;
                autofocusCount=0;
            }
            // Left/Right buttons change autofocus sleep time duration
            else if (padStatus.trigger & nn::hid::CTR::BUTTON_LEFT)
            {
                if (autofocus.SleepTime>0)
                {
                    autofocus.SleepTime-=10;
                    loopCount=0;
                    autofocusCount=0;
                }
            }
            else if (padStatus.trigger & nn::hid::CTR::BUTTON_RIGHT)
            {
                autofocus.SleepTime+=10;
                loopCount=0;
                autofocusCount=0;
            }
        }

        camera.WaitNextImage();

        float floatTx;
        if (autoFocusEnable) // horizontal translation will be computed with autofocus applied
        {
            if (autofocus.IsReadyForNextFrame())
            {
                autofocus.InitImages(camera.GetImageChannel(0),camera.GetImageChannel(1));
                autofocus.NextFrame();
                autofocusCount++;
            }            
            floatTx = autofocus.GetValue()/2.f;
            loopCount++;
        }
        else // horizontal translation will be computed with no autofocus applied
        {
            floatTx = autofocus.GetTranslationX()/2.f;
            autofocus.SetFocus(0.f);
        }

        // This is the amount of translation that will be applied for left(tx) and right(-tx) camera image
        int tx=(floatTx<0.f?floorf(floatTx-0.5f):floorf(floatTx+0.5f));

        // We perform boundary checks for the left camera image translation values
        if ((CAMERA_WIDTH-400)/2 + tx + 400 > CAMERA_WIDTH)
            parallax[0] = CAMERA_WIDTH - 400;
        else if ((CAMERA_WIDTH-400)/2 + tx < 0)
            parallax[0] = 0;
        else
            parallax[0] = (CAMERA_WIDTH-400)/2 + tx;

        // We perform the same boundary checks for the right camera image translation values
        if ((CAMERA_WIDTH-400)/2 - tx + 400 > CAMERA_WIDTH)
            parallax[1] = CAMERA_WIDTH - 400;
        else if ((CAMERA_WIDTH-400)/2 - tx < 0)
            parallax[1] = 0;
        else
            parallax[1] = (CAMERA_WIDTH-400)/2 - tx;

        // Convert camera images from YUV to RGB
        for (u32 i=0; i<2; i++)
        {
            // Start YUV=>RGB conversion in background
            mw::mo::helper::yuvtorgb::LaunchConversion(CAMERA_WIDTH,CAMERA_HEIGHT,nn::y2r::CTR::OUTPUT_RGB_24,nn::y2r::CTR::COEFFICIENT_ITU_R_BT_601,camera.GetImageChannel(i),imageData[i]);
            // During this time do some CPU intensive processing
            fractal.Process();
            // Synchronise with YUV=>RGB conversion
            mw::mo::helper::yuvtorgb::GetEndConversionEvent()->Wait();
        }
        
        camera.PopImage();

        nn::os::Tick time=nn::os::Tick::GetSystemCurrent()-startTime;
        frameCount++;

        // Display some informations on lower screen
        memcpy(mw::mo::helper::displaydevice::GetBackBuffer(NN_GX_DISPLAY1),fractal.GetBuffer(),320*240*3);
        mw::mo::helper::font::PrintLineRgb24((unsigned char*)mw::mo::helper::displaydevice::GetBackBuffer(NN_GX_DISPLAY1), 320, 240, 0,  8*0, 255, 255, 255, "autofocus called ratio: %s%d",autofocusCount?"1/":"",autofocusCount?Round((float)loopCount/(float)autofocusCount):0);
        mw::mo::helper::font::PrintLineRgb24((unsigned char*)mw::mo::helper::displaydevice::GetBackBuffer(NN_GX_DISPLAY1), 320, 240, 0,  8*1, 255, 255, 255, "autofocus sleep time:   %d ms",autofocus.SleepTime);
        mw::mo::helper::font::PrintLineRgb24((unsigned char*)mw::mo::helper::displaydevice::GetBackBuffer(NN_GX_DISPLAY1), 320, 240, 0,  8*2, 255, 255, 255, "fps:                    %d",Round(((float)1000.f*(float)frameCount)/(float)time.ToTimeSpan().GetMilliSeconds()));

        // Display camera images
        const int displayId[2] = {NN_GX_DISPLAY0, NN_GX_DISPLAY0_EXT};
        for (int i=0; i<2; i++)
        {
            int ty;
            if (i == 0)
            {
                float floatTy = autofocus.GetTranslationY();
                ty=(floatTy<0.f?floorf(floatTy-0.5f):floorf(floatTy+0.5f));
                
                // We perform boundary checks for the left camera image horizontal translation value
                if ((CAMERA_HEIGHT-240)/2 + ty + 240 > CAMERA_HEIGHT)
                    ty = CAMERA_HEIGHT - 240 - (CAMERA_HEIGHT-240)/2;
                else if ((CAMERA_HEIGHT-240)/2 + ty < 0)
                    ty = - (CAMERA_HEIGHT-240)/2;
            }
            else // No translation performed for right camera image
                ty = 0;

            image[i].SetImageSelection(parallax[i], (CAMERA_HEIGHT-240)/2 + ty, 400, 240);

            mw::mo::helper::displaydevice::BeginDraw(displayId[i]);
            mw::mo::helper::displaydevice::Clear(0.f,0.f,0.f,0.f);
            image[i].SetTransparency(1.f);
            image[i].Draw();            
            mw::mo::helper::displaydevice::EndDraw();
        }        
        mw::mo::helper::displaydevice::CommitDraw();
        mw::mo::helper::displaydevice::Flip(NN_GX_DISPLAY_BOTH);        
        
        // Reset the fps timer every 2 seconds
        if (time.ToTimeSpan().GetSeconds()>=2)
        {
            startTime=nn::os::Tick::GetSystemCurrent();
            frameCount=0;
        }

        if (mw::mo::helper::application::ProcessStatus() == true)
            break;
    }

    /*-------------------------------------------------------------------------------------------------------*
      Releases all memory and objects used before ending application.
     *-------------------------------------------------------------------------------------------------------*/

    // Synchronize with autofocus thread
    autofocus.Synchronize();    

    // Detach and destroy thread and associated object
    autofocusThread.Detach();
    autofocusThread.Finalize();    
    autofocus.Finalize();

    // Release fractal object
    fractal.Finalize();    

    // End camera capture and release object
    camera.Finalize();

    // Free allocated display buffers
    for (u32 i=0; i<2; i++)
    {
        image[i].Finalize();
        s_Heap.Free(imageData[i]);
    }    

    mw::mo::helper::yuvtorgb::Finalize();
    mw::mo::helper::displaydevice::Finalize();

    // Dump for memory leak
    s_Heap.Dump();

    /*-------------------------------------------------------------------------------------------------------*
      End of application.
     *-------------------------------------------------------------------------------------------------------*/

    NN_LOG("Application ended...\n");

    mw::mo::helper::application::Finalize();

    nn::applet::PrepareToCloseApplication();
    nn::applet::CloseApplication();

    /*-------------------------------------------------------------------------------------------------------*/
}

/*-----------------------------------------------------------------------------------------------------------*
  EOF
*------------------------------------------------------------------------------------------------------------*/

