/*-----------------------------------------------------------------------------------------------------------*
 File:           helper_Camera.h
 Description:    Helper file that manages camera capture

 Copyright (C)2011 Mobiclip, Inc.
 *-----------------------------------------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/fnd.h>
#include <string.h>
#include <nn/nstd/nstd_Printf.h>

/*-----------------------------------------------------------------------------------------------------------*/

#include "helper_Camera.h"

/*-----------------------------------------------------------------------------------------------------------*/

using namespace mw::mo::helper;

/*-----------------------------------------------------------------------------------------------------------*
  Constants definition.
 *-----------------------------------------------------------------------------------------------------------*/

const int CAPTURE_THREAD_PRIORITY   = 3;
const int CAPTURE_THREAD_STACK_SIZE = 4096;

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

void Camera::CaptureThreadFunc (Camera* camera)
{
    for(;;)
    {
        /*---------------------------------------------------------------------------------------------------*/

        s32 eventIndex = nn::os::WaitObject::WaitAny(reinterpret_cast<nn::os::WaitObject**>(camera->m_EventArray), camera->m_NbEvents);

        /*---------------------------------------------------------------------------------------------------*/

        if (eventIndex == 0)
        {
            camera->m_EndCaptureEvent.ClearSignal();
            
            nn::camera::CTR::StopCapture(nn::camera::CTR::PORT_CAM1);

            nn::camera::CTR::ClearBuffer(nn::camera::CTR::PORT_CAM1);

            if (camera->m_NbVideoImage == 2)
            {
                nn::camera::CTR::StopCapture(nn::camera::CTR::PORT_CAM2);

                nn::camera::CTR::ClearBuffer(nn::camera::CTR::PORT_CAM2);
            }
    
            return;
        }

        /*---------------------------------------------------------------------------------------------------*/

        if ((camera->m_NbVideoImage == 2 && (eventIndex == 3 || eventIndex == 4)) ||
            (camera->m_NbVideoImage == 1 && (eventIndex == 2))) //Transfert error
        {
            //NN_LOG("Camera capture error...\n");

            nn::camera::CTR::StopCapture(nn::camera::CTR::PORT_CAM1);
            nn::camera::CTR::ClearBuffer(nn::camera::CTR::PORT_CAM1);

            if (camera->m_NbVideoImage == 2)
            {
                nn::camera::CTR::StopCapture(nn::camera::CTR::PORT_CAM2);
                nn::camera::CTR::ClearBuffer(nn::camera::CTR::PORT_CAM2);
            }
            
            nn::camera::CTR::SetReceiving(&camera->m_ReceiveEvent[0],
                                              reinterpret_cast<void*>(camera->m_Yuv422VideoImageData[0][camera->m_CapturedFrameIndex[0]%NB_CAMERA_BUFFER]), 
                                              nn::camera::CTR::PORT_CAM1, camera->m_CameraOutputSize, camera->m_TransferUnit);

            if (camera->m_NbVideoImage == 2)
                nn::camera::CTR::SetReceiving(&camera->m_ReceiveEvent[1],
                                              reinterpret_cast<void*>(camera->m_Yuv422VideoImageData[1][camera->m_CapturedFrameIndex[1]%NB_CAMERA_BUFFER]), 
                                              nn::camera::CTR::PORT_CAM2, camera->m_CameraOutputSize, camera->m_TransferUnit);

            nn::camera::CTR::Port port;
            if (camera->m_NbVideoImage == 2)
                port = nn::camera::CTR::PORT_BOTH;
            else
                port = nn::camera::CTR::PORT_CAM1;

            nn::camera::CTR::StartCapture(port);
        }
        else
        {
            int camIndex = eventIndex-1;
            if ((camera->m_CapturedFrameIndex[camIndex] - camera->m_RetrievedFrameIndex[camIndex]) < (NB_CAMERA_BUFFER-1))
                camera->m_CapturedFrameIndex[camIndex]++;

            camera->m_ClientEvent.Signal(); // Wake up the waiting thread
            
            if (camIndex == 0)
                nn::camera::CTR::SetReceiving(&camera->m_ReceiveEvent[camIndex],reinterpret_cast<void*>(camera->m_Yuv422VideoImageData[0][camera->m_CapturedFrameIndex[camIndex]%NB_CAMERA_BUFFER]), nn::camera::CTR::PORT_CAM1, camera->m_CameraOutputSize, camera->m_TransferUnit);
            else
                nn::camera::CTR::SetReceiving(&camera->m_ReceiveEvent[camIndex],reinterpret_cast<void*>(camera->m_Yuv422VideoImageData[1][camera->m_CapturedFrameIndex[camIndex]%NB_CAMERA_BUFFER]), nn::camera::CTR::PORT_CAM2, camera->m_CameraOutputSize, camera->m_TransferUnit);
        }

        /*---------------------------------------------------------------------------------------------------*/
    }
}

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

bool Camera::Initialize (nn::fnd::IAllocator* pAllocator, 
                         nn::camera::CTR::CameraSelect cameraSelect, 
                         nn::camera::CTR::Size size, 
                         nn::camera::CTR::FrameRate framerate)
{
    int width;
    int height;

    switch(size)
    {
    case nn::camera::CTR::SIZE_QQVGA:
        width = 160;
        height = 120;
        break;
    case nn::camera::CTR::SIZE_QCIF:
        width = 176;
        height = 144;
        break;
    case nn::camera::CTR::SIZE_DS_LCD:
        width = 256;
        height = 192;
        break;
    case nn::camera::CTR::SIZE_QVGA:
        width = 320;
        height = 240;
        break;
    case nn::camera::CTR::SIZE_CTR_TOP_LCD:
        width = 400;
        height = 240;
        break;
    case nn::camera::CTR::SIZE_CIF:
        width = 352;
        height = 288;
        break;
    case nn::camera::CTR::SIZE_DS_LCDx4:
        width = 512;
        height = 384;
        break;
    case nn::camera::CTR::SIZE_VGA:
        width = 640;
        height = 480;
        break;
    default:
        NN_LOG("Camera::Initialize: unknown camera size\n");
        return false;
    }

    return Initialize(pAllocator, cameraSelect, framerate, width, height);

}

bool Camera::Initialize (nn::fnd::IAllocator* pAllocator,
                         nn::camera::CTR::CameraSelect cameraSelect,
                         nn::camera::CTR::FrameRate framerate,
                         int width, int height,
                         int selectionPosX, int selectionPosY,
                         int selectionWidth, int selectionHeight)
{
    /*-------------------------------------------------------------------------------------------------------*/

    StopCapture();

    nn::camera::CTR::Initialize();
    
    /*-------------------------------------------------------------------------------------------------------*/

    m_NbEvents = 0;

    m_EndCaptureEvent.Initialize(true);
    m_EventArray[m_NbEvents++] = &m_EndCaptureEvent;
    
    /*-------------------------------------------------------------------------------------------------------*/

    m_Allocator = pAllocator;

    /*-------------------------------------------------------------------------------------------------------*/

    m_Width = width>640 ? 640 : width;
    m_Height = height>480 ? 480 : height;

    if (   selectionPosX >= 0
        && selectionPosY >= 0
        && selectionWidth > 0
        && selectionHeight > 0)
    {
        m_SelectionPosX = selectionPosX;
        m_SelectionPosY = selectionPosY;
        m_SelectionWidth = selectionWidth;
        m_SelectionHeight = selectionHeight;
    }
    else
    {
        // Compute a best fit crop in the 640x480 capture buffer
        if (m_Width*480 > m_Height*640)
        {
            m_SelectionWidth = 640;
            m_SelectionHeight = 640*m_Height/m_Width;
            m_SelectionPosX = 0;
            m_SelectionPosY = (480 - m_SelectionHeight + 1)>>1;
        }
        else
        {
            m_SelectionWidth = 480*m_Width/m_Height;
            m_SelectionHeight = 480;
            m_SelectionPosX = (640 - m_SelectionWidth + 1)>>1;
            m_SelectionPosY = 0;
        }

        // Security
        m_SelectionWidth = m_SelectionWidth>640 ? 640 : m_SelectionWidth;
        m_SelectionHeight = m_SelectionHeight>480 ? 480 : m_SelectionHeight;
    }

    /*-------------------------------------------------------------------------------------------------------*/

    nn::camera::CTR::Port port;
    if (cameraSelect == nn::camera::CTR::SELECT_OUT1_OUT2)
    {
        port = nn::camera::CTR::PORT_BOTH;
        m_NbVideoImage = 2;
    }
    else
    {
        port = nn::camera::CTR::PORT_CAM1;
        m_NbVideoImage = 1;
    }

    /*-------------------------------------------------------------------------------------------------------*/

    s16 transferLine = nn::camera::CTR::GetMaxLines(m_Width, m_Height);

    nn::camera::CTR::SetTransferLines(port, transferLine, m_Width, m_Height);

    /*-------------------------------------------------------------------------------------------------------*/
    
    if (cameraSelect == nn::camera::CTR::SELECT_OUT1_OUT2)
    {
        nn::camera::CTR::GetBufferErrorInterruptEvent(&m_ErrorEvent[0], nn::camera::CTR::PORT_CAM1);
        nn::camera::CTR::GetBufferErrorInterruptEvent(&m_ErrorEvent[1], nn::camera::CTR::PORT_CAM2);
        nn::camera::CTR::GetVsyncInterruptEvent(&m_VSyncEvent[0],nn::camera::CTR::PORT_CAM1);
        nn::camera::CTR::GetVsyncInterruptEvent(&m_VSyncEvent[1],nn::camera::CTR::PORT_CAM2);
        m_EventArray[m_NbEvents++] = &m_ReceiveEvent[0];
        m_EventArray[m_NbEvents++] = &m_ReceiveEvent[1];
        m_EventArray[m_NbEvents++] = &m_ErrorEvent[0];
        m_EventArray[m_NbEvents++] = &m_ErrorEvent[1];
    }
    else
    {
        nn::camera::CTR::GetBufferErrorInterruptEvent(&m_ErrorEvent[0], nn::camera::CTR::PORT_CAM1);
        nn::camera::CTR::GetVsyncInterruptEvent(&m_VSyncEvent[0],nn::camera::CTR::PORT_CAM1);
        m_EventArray[m_NbEvents++] = &m_ReceiveEvent[0];
        m_EventArray[m_NbEvents++] = &m_ErrorEvent[0];
    }

    /*-------------------------------------------------------------------------------------------------------*/

    for (int i=0; i<m_NbVideoImage; i++)
    {
        for (int j=0; j<NB_CAMERA_BUFFER; j++)
        {
            m_Yuv422VideoImageData[i][j] = reinterpret_cast<u8*>(m_Allocator->Allocate(m_Width * m_Height * 2, 128));
            if (m_Yuv422VideoImageData[i][j] == NULL)
            {
                NN_LOG("Camera::Initialize: can't allocate YUV422 image\n");
                return false;
            }
        }
    }

    /*-------------------------------------------------------------------------------------------------------*/

    nn::camera::CTR::SetFrameRate(nn::camera::CTR::SELECT_ALL, framerate);

    nn::Result res;
    if( m_NbVideoImage == 2 )
    {
        res = nn::camera::CTR::SwitchContext(nn::camera::CTR::SELECT_OUT1,nn::camera::CTR::CONTEXT_A);
        if (res.IsFailure())
        {
            NN_LOG("Camera::Initialize: can't switch context for camera\n");
            return false;
        }
        res = nn::camera::CTR::SwitchContext(nn::camera::CTR::SELECT_OUT2,nn::camera::CTR::CONTEXT_B);
        if (res.IsFailure())
        {
            NN_LOG("Camera::Initialize: can't switch context for camera\n");
            return false;
        }
        res = nn::camera::CTR::SetDetailSize(nn::camera::CTR::SELECT_OUT1, m_Width, m_Height, m_SelectionPosX, m_SelectionPosY, m_SelectionPosX + m_SelectionWidth - 1, m_SelectionPosY + m_SelectionHeight - 1, nn::camera::CTR::CONTEXT_A);
        if (res.IsFailure())
        {
            NN_LOG("Camera::Initialize: can't set detailsize for camera\n");
            return false;
        }
        res = nn::camera::CTR::SetDetailSize(nn::camera::CTR::SELECT_OUT2, m_Width, m_Height, m_SelectionPosX, m_SelectionPosY, m_SelectionPosX + m_SelectionWidth - 1, m_SelectionPosY + m_SelectionHeight - 1, nn::camera::CTR::CONTEXT_B);
        if (res.IsFailure())
        {
            NN_LOG("Camera::Initialize: can't set detailsize for camera\n");
            return false;
        }
    }
    else
    {
        res = nn::camera::CTR::SwitchContext(cameraSelect,nn::camera::CTR::CONTEXT_A);
        if (res.IsFailure())
        {
            NN_LOG("Camera::Initialize: cant switch context for camera\n");
            return false;
        }
        res = nn::camera::CTR::SetDetailSize(cameraSelect, m_Width, m_Height, m_SelectionPosX, m_SelectionPosY, m_SelectionPosX + m_SelectionWidth - 1, m_SelectionPosY + m_SelectionHeight - 1, nn::camera::CTR::CONTEXT_A);
        if (res.IsFailure())
        {
            NN_LOG("Camera::Initialize: can't set detailsize for camera\n");
            return false;
        }
    }

    res = nn::camera::CTR::Activate(cameraSelect);
    if(res == nn::camera::CTR::ResultFatalError())
    {
        NN_LOG("Camera::Initialize: can't activate camera\n");
        return false;
    }

    /*-------------------------------------------------------------------------------------------------------*/

    m_ClientEvent.Initialize( false );

    /*-------------------------------------------------------------------------------------------------------*/

    return true;

    /*-------------------------------------------------------------------------------------------------------*/
}

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

void Camera::Finalize(void)
{
    /*-------------------------------------------------------------------------------------------------------*/

    StopCapture();

    for (int i=0; i<m_NbVideoImage; i++)
    {
        for (int j=0; j<NB_CAMERA_BUFFER; j++)
        {    
            m_Allocator->Free(reinterpret_cast<void*>(m_Yuv422VideoImageData[i][j]));
            m_Yuv422VideoImageData[i][j] = 0;
        }
    }

    m_NbVideoImage = 0;

    m_EndCaptureEvent.Finalize();

    nn::camera::CTR::Finalize();

    m_ClientEvent.Finalize();

    /*-------------------------------------------------------------------------------------------------------*/
}

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

bool Camera::StartCapture (void)
{
    /*-------------------------------------------------------------------------------------------------------*/
    
    if (m_Capturing == true)
        return false;

    for (int i=0; i< m_NbVideoImage; i++)
    {
        m_CapturedFrameIndex[i] = 0;
        m_RetrievedFrameIndex[i] = 0;
    }

    /*-------------------------------------------------------------------------------------------------------*/

    m_TransferUnit = nn::camera::CTR::GetTransferBytes(nn::camera::CTR::PORT_CAM1);

    m_CameraOutputSize = nn::camera::CTR::GetFrameBytes(m_Width, m_Height);
    
    nn::camera::CTR::SetReceiving(&m_ReceiveEvent[0],
                                  reinterpret_cast<void*>(m_Yuv422VideoImageData[0][m_CapturedFrameIndex[0]]), 
                                  nn::camera::CTR::PORT_CAM1, m_CameraOutputSize, m_TransferUnit);
    
    if (m_NbVideoImage == 2)
        nn::camera::CTR::SetReceiving(&m_ReceiveEvent[1],
                                      reinterpret_cast<void*>(m_Yuv422VideoImageData[1][m_CapturedFrameIndex[1]]), 
                                      nn::camera::CTR::PORT_CAM2, m_CameraOutputSize, m_TransferUnit);


    nn::camera::CTR::Port port;
    if (m_NbVideoImage == 2)
        port = nn::camera::CTR::PORT_BOTH;
    else
        port = nn::camera::CTR::PORT_CAM1;

    nn::camera::CTR::ClearBuffer(port);

    /*-------------------------------------------------------------------------------------------------------*/

    nn::camera::CTR::StartCapture(port);

    m_CaptureThread.StartUsingAutoStack(CaptureThreadFunc, this, CAPTURE_THREAD_STACK_SIZE, CAPTURE_THREAD_PRIORITY);

    m_Capturing = true;

    return true;
    
    /*-------------------------------------------------------------------------------------------------------*/
}

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

bool Camera::StopCapture (void)
{
    /*-------------------------------------------------------------------------------------------------------*/

    if (m_Capturing == false)
        return false;
    
    m_EndCaptureEvent.Signal();

    m_CaptureThread.Join();
    m_CaptureThread.Finalize();

    m_Capturing = false;

    return true;

    /*-------------------------------------------------------------------------------------------------------*/
}

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

bool Camera::IsImageAvailable (void)
{
    /*-------------------------------------------------------------------------------------------------------*/

    if ((m_CapturedFrameIndex[0] - m_RetrievedFrameIndex[0]) == 0)
        return false;

    if (m_NbVideoImage == 2)
        if ((m_CapturedFrameIndex[1] - m_RetrievedFrameIndex[1]) == 0)
            return false;

    return true;

    /*-------------------------------------------------------------------------------------------------------*/
}

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

void Camera::WaitNextImage(void)
{
    /*-------------------------------------------------------------------------------------------------------*/

    while( ((m_CapturedFrameIndex[0] - m_RetrievedFrameIndex[0]) == 0) || 
           ( (m_NbVideoImage == 2) && ((m_CapturedFrameIndex[1] - m_RetrievedFrameIndex[1]) == 0) ) )
    {
        m_ClientEvent.Wait();
    }

    /*-------------------------------------------------------------------------------------------------------*/
}

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

void Camera::PopImage (void)
{
    if (IsImageAvailable())
    {
        m_RetrievedFrameIndex[0]++;    
        if (m_NbVideoImage == 2)
            m_RetrievedFrameIndex[1]++;    
    }
}

/*-----------------------------------------------------------------------------------------------------------*

 *-----------------------------------------------------------------------------------------------------------*/

void Camera::ChangeResolution(nn::camera::CameraSelect cameraSelect,int width, int height)
{
    // Stereo images not supported yet
    if (m_NbVideoImage == 2)
        return;

    m_Width = width>640 ? 640 : width;
    m_Height = height>480 ? 480 : height;
    
    if (m_Width*480 > m_Height*640)
    {
        m_SelectionWidth = 640;
        m_SelectionHeight = 640*m_Height/m_Width;
        m_SelectionPosX = 0;
        m_SelectionPosY = (480 - m_SelectionHeight + 1)>>1;
    }
    else
    {
        m_SelectionWidth = 480*m_Width/m_Height;
        m_SelectionHeight = 480;
        m_SelectionPosX = (640 - m_SelectionWidth + 1)>>1;
        m_SelectionPosY = 0;
    }

    m_SelectionWidth = m_SelectionWidth>640 ? 640 : m_SelectionWidth;
    m_SelectionHeight = m_SelectionHeight>480 ? 480 : m_SelectionHeight;

    m_EndCaptureEvent.Signal();
    m_CaptureThread.Join();
    m_CaptureThread.Finalize();
    while (nn::camera::IsBusy(nn::camera::PORT_CAM1)) {}

    m_VSyncEvent[0].Wait();

    nn::camera::SetTransferLines(nn::camera::PORT_CAM1,nn::camera::GetMaxLines(m_Width,m_Height),m_Width,m_Height);

    nn::camera::SetDetailSize(cameraSelect, m_Width, m_Height, m_SelectionPosX, m_SelectionPosY, m_SelectionPosX + m_SelectionWidth - 1, m_SelectionPosY + m_SelectionHeight - 1, nn::camera::CONTEXT_A);

    m_TransferUnit = nn::camera::GetTransferBytes(nn::camera::PORT_CAM1);

    m_CameraOutputSize = nn::camera::GetFrameBytes(m_Width, m_Height);

    m_CapturedFrameIndex[0] = 0;
    m_RetrievedFrameIndex[0] = 0;
    
    nn::camera::SetReceiving(&m_ReceiveEvent[0],
                             reinterpret_cast<void*>(m_Yuv422VideoImageData[0][m_CapturedFrameIndex[0]]), 
                             nn::camera::PORT_CAM1, m_CameraOutputSize, m_TransferUnit);

    nn::camera::StartCapture(nn::camera::PORT_CAM1);

    m_CaptureThread.StartUsingAutoStack(CaptureThreadFunc, this, CAPTURE_THREAD_STACK_SIZE, CAPTURE_THREAD_PRIORITY);
}

